# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from __future__ import annotations

import uuid
from pathlib import Path

import numpy as np
import onnx
from onnx import TensorProto
from onnx.helper import float32_to_float8e4m3, np_dtype_to_tensor_dtype
from onnx.numpy_helper import float8e4m3_to_float32
from onnx.reference import ReferenceEvaluator
from onnx.reference import ops as onnx_ops
from onnx.reference.custom_element_types import float8e4m3fn, float8e4m3fnuz, float8e5m2, float8e5m2fnuz
from onnx.reference.op_run import OpRun

import onnxruntime
import onnxruntime.capi._pybind_state as C
from onnxruntime.quantization import CalibrationDataReader

onnx_recent_enough = hasattr(OpRun, "infer_name")

if onnx_recent_enough:
    # Test with ReferenceEvaluator requires PR https://github.com/onnx/onnx/pull/5408/.
    # https://github.com/onnx/onnx/pull/5408
    try:
        from onnx.reference.op_run import to_array_extended

    except ImportError:
        to_array_extended = None
        onnx_recent_enough = False


class QOpRun(OpRun):
    op_domain = "com.microsoft"

    f8_types = {
        TensorProto.FLOAT8E4M3FN,
        TensorProto.FLOAT8E4M3FNUZ,
        TensorProto.FLOAT8E5M2,
        TensorProto.FLOAT8E5M2FNUZ,
    }

    def get_tensor_type(self, tensor: np.ndarray) -> int:
        if tensor.dtype == float8e4m3fn and tensor.dtype.descr[0][0] == "e4m3fn":
            return TensorProto.FLOAT8E4M3FN
        if tensor.dtype == float8e4m3fnuz and tensor.dtype.descr[0][0] == "e4m3fnuz":
            return TensorProto.FLOAT8E4M3FNUZ
        if tensor.dtype == float8e5m2 and tensor.dtype.descr[0][0] == "e5m2":
            return TensorProto.FLOAT8E5M2
        if tensor.dtype == float8e5m2fnuz and tensor.dtype.descr[0][0] == "e5m2fnuz":
            return TensorProto.FLOAT8E5M2FNUZ
        return np_dtype_to_tensor_dtype(tensor.dtype)


class QGemm(QOpRun):
    def _run(
        self,
        A,
        a_scale,
        a_zero_point,
        B,
        b_scale,
        b_zero_point,
        C=None,
        y_scale=None,
        y_zero_point=None,
        transA=0,
        transB=0,
        alpha=1.0,
    ):
        if transA:
            A = A.T
        if transB:
            B = B.T

        a_type = self.get_tensor_type(a_zero_point)
        b_type = self.get_tensor_type(b_zero_point)
        y_type = self.get_tensor_type(y_zero_point)
        if a_type == TensorProto.FLOAT8E4M3FN and b_type == TensorProto.FLOAT8E4M3FN:
            a_scaled = (float8e4m3_to_float32(A).astype(float) - float8e4m3_to_float32(a_zero_point)) * np.float32(
                a_scale
            )
            b_scaled = (float8e4m3_to_float32(B).astype(float) - float8e4m3_to_float32(b_zero_point)) * np.float32(
                b_scale
            )
            y = a_scaled @ b_scaled * np.float32(alpha)
            if C is not None:
                dtype = self.get_tensor_type(C)
                if dtype not in (TensorProto.FLOAT, TensorProto.FLOAT16):
                    raise TypeError(f"C.dtype must be float16 or float 32 not {dtype}.")
                y += C.astype(np.float32)
            if y_scale is not None:
                y /= y_scale
            if y_zero_point is not None:
                y += float8e4m3_to_float32(y_zero_point)
                ry = y.ravel()

                fy = np.empty(ry.shape, dtype=float8e4m3fn)
                for i in range(fy.shape[0]):
                    el = float32_to_float8e4m3(ry[i])  # type: ignore[assignment]
                    fy[i] = el
                y = fy.reshape(y.shape)
            else:
                raise NotImplementedError("y_zero_point is not empty. QGemm is not implemented in that case.")
            return (y,)
        elif a_type in self.f8_types or b_type in self.f8_types or y_type in self.f8_types:
            raise NotImplementedError(f"QGemm not implemented for zero_types {a_type}, {b_type}, {y_type}.")
        else:
            if TensorProto.FLOAT8E4M3FN in {a_type, b_type, y_type}:
                raise TypeError(f"Unexpected type for A: {dtype}, B:{dtype} or Y:{dtype}.")
            a_scaled = (A.astype(float) - a_zero_point) * np.float32(a_scale)
            b_scaled = (B.astype(float) - b_zero_point) * np.float32(b_scale)
            y = a_scaled @ b_scaled * np.float32(alpha)
            if C is not None:
                y += C * np.float32(a_scale) * np.float32(b_scale)
            if y_scale is not None:
                y /= np.float32(y_scale)
            if y_zero_point is not None:
                y += y_zero_point

            if y_zero_point is not None:
                dtype = y_zero_point.dtype
            elif C is not None:
                dtype = C.dtype
            else:
                dtype = A.dtype

            y = np.rint(y)
            if dtype == np.uint8:
                y = np.clip(y, 0, 255)
            elif dtype == np.int8:
                y = np.clip(y, -128, 127)
            else:
                raise ValueError(f"Unexpected dtype={dtype}, it should be uint8 or int8.")

            return (y.astype(dtype),)


class QLinearMatMul(QOpRun):
    def _run(
        self,
        A,
        a_scale,
        a_zero_point,
        B,
        b_scale,
        b_zero_point,
        y_scale=None,
        y_zero_point=None,
    ):
        a_type = self.get_tensor_type(a_zero_point)
        b_type = self.get_tensor_type(b_zero_point)
        y_type = self.get_tensor_type(y_zero_point)
        if a_type == TensorProto.FLOAT8E4M3FN and b_type == TensorProto.FLOAT8E4M3FN:
            a_scaled = (float8e4m3_to_float32(A).astype(float) - float8e4m3_to_float32(a_zero_point)) * np.float32(
                a_scale
            )
            b_scaled = (float8e4m3_to_float32(B).astype(float) - float8e4m3_to_float32(b_zero_point)) * np.float32(
                b_scale
            )
            y = a_scaled @ b_scaled
            if y_scale is not None:
                y /= y_scale
            if y_zero_point is not None:
                y += float8e4m3_to_float32(y_zero_point)
                ry = y.ravel()

                fy = np.empty(ry.shape, dtype=float8e4m3fn)
                for i in range(fy.shape[0]):
                    el = float32_to_float8e4m3(ry[i])  # type: ignore[assignment]
                    fy[i] = el
                y = fy.reshape(y.shape)
            else:
                raise NotImplementedError("y_zero_point is not empty. QLinearMatMul is not implemented in that case.")
            return (y,)
        elif a_type in self.f8_types or b_type in self.f8_types or y_type in self.f8_types:
            raise NotImplementedError(f"QLinearMatMul not implemented for zero_types {a_type}, {b_type}, {y_type}.")
        else:
            if TensorProto.FLOAT8E4M3FN in {a_type, b_type, y_type}:
                raise TypeError(f"Unexpected type for A: {a_type}, B:{b_type} or Y:{y_type}.")
            a_scaled = (A.astype(float) - a_zero_point) * np.float32(a_scale)
            b_scaled = (B.astype(float) - b_zero_point) * np.float32(b_scale)
            y = a_scaled @ b_scaled
            if y_scale is not None:
                y /= np.float32(y_scale)
            if y_zero_point is not None:
                y += y_zero_point

            if y_zero_point is not None:
                dtype = y_zero_point.dtype
            elif C is not None:
                dtype = C.dtype
            else:
                dtype = A.dtype

            y = np.rint(y)
            if dtype == np.uint8:
                y = np.clip(y, 0, 255)
            elif dtype == np.int8:
                y = np.clip(y, -128, 127)
            else:
                raise ValueError(f"Unexpected dtype={dtype}, it should be uint8 or int8.")

            return (y.astype(dtype),)


class TestDataFeeds(CalibrationDataReader):
    def __init__(self, data_feeds):
        """
        parameter data_feeds: list of input feed, each input feed is diction of {input_name: np_array}
        """
        self.data_feeds = data_feeds
        self.iter_next = iter(self.data_feeds)

    def get_next(self):
        return next(self.iter_next, None)

    def rewind(self):
        self.iter_next = iter(self.data_feeds)


def input_feeds_neg_one_zero_one(n, name2shape, seed=None):
    """
    randomize n feed according to shape, its values are from -1, 0, and 1
    """
    if seed is not None:
        np.random.seed(seed)

    input_data_list = []
    for _i in range(n):
        inputs = {}
        for name, shape in name2shape.items():
            inputs.update({name: np.random.randint(-1, 2, shape).astype(np.float32)})
        input_data_list.extend([inputs])
    dr = TestDataFeeds(input_data_list)
    return dr


def input_feeds_neg_one_zero_one_list(n, name2shape, seed=None):
    """
    randomize n feed according to shape, its values are from -1, 0, and 1
    """
    if seed is not None:
        np.random.seed(seed)

    input_data_list = []
    for _i in range(n):
        inputs = {}
        for name, shape in name2shape.items():
            inputs.update({name: np.random.randint(-1, 2, shape).astype(np.float32)})
        input_data_list.extend([inputs])
    return input_data_list


class GenerateCalibrationData(CalibrationDataReader):
    def __init__(self, data_list, input_nodes, input_shapes, no_tensor_num, in_dtypes, inputs_conv_channel_last=None):
        print("Generating calibration dataset from " + str(data_list))
        print("input nodes are ", input_nodes, "input shapes are ", input_shapes)
        if inputs_conv_channel_last:
            print(f"Inputs that will be converted to channel last: {inputs_conv_channel_last}")

        self.enum_data_dicts = []
        self.input_nodes = input_nodes
        self.input_shapes = input_shapes
        self.inputs_conv_channel_last = inputs_conv_channel_last
        self.calibration_dataset = data_list

    def __len__(self):
        return len(self.calibration_dataset)

    def get_next(self):
        feed_dict = {}
        inp = next(self.calibration_dataset, None)
        if inp is not None:
            for i in range(len(self.input_nodes)):
                input_data = inp[i].reshape(self.input_shapes[i])
                if self.inputs_conv_channel_last is not None and self.input_nodes[i] in self.inputs_conv_channel_last:
                    input_data = np.moveaxis(input_data, 1, -1)
                dict_item = {self.input_nodes[i]: input_data}
                feed_dict.update(dict_item)
            return feed_dict
        else:
            return None


class StridedDataReader(GenerateCalibrationData):
    def __init__(
        self,
        data_list,
        input_nodes,
        input_shapes,
        no_tensor_num,
        in_dtypes,
        inputs_conv_channel_last=None,
        stride=1,
        start_index=0,
        end_index=None,
    ):
        super().__init__(data_list, input_nodes, input_shapes, no_tensor_num, in_dtypes, inputs_conv_channel_last)

        self.stride = max(1, stride)  # Ensure stride is at least 1
        self.start_index = start_index
        self.end_index = (
            end_index if end_index is not None else len(self.calibration_dataset)
        )  # Default to the end of the dataset
        self.enum_data_dicts = iter([])

    def get_next(self):
        iter_data = next(self.enum_data_dicts, None)
        if iter_data:
            return iter_data

        self.enum_data_dicts = None
        if self.start_index < self.end_index:
            print(f"start index is {self.start_index}")
            data = self.load_serial()

            self.start_index += self.stride
            self.enum_data_dicts = iter(data)

            return next(self.enum_data_dicts, None)
        else:
            return None

    def load_serial(self):
        batch_data = []
        end_loop = min(self.end_index, self.start_index + self.stride)
        for i in range(self.start_index, end_loop):
            print(f"debugging the load serial index {i}")
            data_item = self.calibration_dataset[i]
            processed_item = self.process_data_item(data_item)
            batch_data.append(processed_item)
        return batch_data

    def process_data_item(self, data_item):
        feed_dict = {}
        for _, node in enumerate(self.input_nodes):
            # input_data = data_item[i].reshape(self.input_shapes[i])
            feed_dict[node] = data_item["input"]
        return feed_dict

    def set_range(self, start_index, end_index=None):
        self.start_index = start_index
        self.end_index = end_index if end_index is not None else len(self.calibration_dataset)
        self.enum_data_dicts = iter([])

    def rewind(self):
        """Rewind the data reader to the beginning of the dataset."""
        self.start_index = 0
        self.enum_data_dicts = iter([])


def check_op_type_order(testcase, model_to_check, ops):
    if isinstance(model_to_check, str):
        model = onnx.load(model_to_check)
    elif isinstance(model_to_check, onnx.ModelProto):
        model = model_to_check

    testcase.assertEqual(len(ops), len(model.graph.node), "op count is not same")
    for node_idx, node in enumerate(model.graph.node):
        testcase.assertEqual(
            ops[node_idx],
            node.op_type,
            f"op {node_idx} is not in order. Expected: {ops[node_idx]}, Actual: {node.op_type}",
        )


def check_op_type_count(testcase, model_path, **kwargs):
    model = onnx.load(Path(model_path))
    optype2count = {}
    for op_type in kwargs:
        optype2count[op_type] = 0
    for node in model.graph.node:
        if node.op_type in optype2count:
            optype2count[node.op_type] += 1

    for op_type, value in kwargs.items():
        try:
            testcase.assertEqual(
                value,
                optype2count[op_type],
                f"op_type {op_type} count not same",
            )
        except AssertionError as e:
            from onnx_array_api.plotting.text_plot import onnx_simple_text_plot

            raise AssertionError(
                f"Assert failed:\noptype={optype2count}\nkwargs={kwargs}\n{onnx_simple_text_plot(model)}"
            ) from e


def check_sign_f8_quantization(model_path_origin, model_path_to_check):
    """
    Quantization to float 8 type does not change the sign as zero_point is always null.
    This function checks that the quantized parameters did not change.
    """
    with open(model_path_origin, "rb") as f:
        model = onnx.load(f)
    names = {init.name: init for init in model.graph.initializer}
    with open(model_path_to_check, "rb") as f:
        model_f8 = onnx.load(f)
    names_f8 = {init.name: init for init in model_f8.graph.initializer}
    for init in model_f8.graph.initializer:
        if not init.name.endswith("_quantized"):
            continue
        name = init.name.replace("_quantized", "")
        if name not in names:
            raise AssertionError(f"Unable to find {name!r} in {set(names)}.")
        scale_zp = [i.name for i in model_f8.graph.initializer if i.name.startswith(name)]
        if len(scale_zp) not in (1, 3):
            raise AssertionError(
                f"Need one or three names not {scale_zp}, all names: { {i.name for i in model_f8.graph.initializer} }."
            )
        scale = [name for name in scale_zp if "scale" in name]
        zero = [name for name in scale_zp if "zero" in name]
        if len(scale_zp) == 3:
            if len(scale) != 1:
                raise AssertionError(f"Need one name not {scale}.")
            if len(zero) != 1:
                raise AssertionError(f"Need one name not {zero}.")
        else:
            if len(scale) != 0:
                raise AssertionError(f"No scale is expected but has {scale}.")
            if len(zero) != 0:
                raise AssertionError(f"No zero is expected but has {zero}.")

        expected_sign = onnx.numpy_helper.to_array(names[name]) >= 0

        if "bias" in init.name:
            if init.data_type >= 17:
                raise AssertionError(f"bias {init.name!r} should be float 16 not {init.data_type}.")
            continue
        if init.data_type < 17:
            raise AssertionError(f"Initializer {init.name!r} not a float 8 type.")
        raw = np.array([int(i) for i in init.raw_data])
        got_sign = raw <= 128
        try:
            np.testing.assert_allclose(expected_sign.ravel(), got_sign)
        except AssertionError as e:
            scale_value = onnx.numpy_helper.to_array(names_f8[scale[0]])
            err_msg = f"Sign are different for {name!r}, scale={scale_value}."
            if to_array_extended is not None:
                values = onnx.numpy_helper.to_array(names[name]).flatten()
                f8_values = to_array_extended(init)
                zero = onnx_ops.op_cast.Cast_19.eval(np.array(0), to=init.data_type)
                dq = onnx_ops.op_dequantize_linear.DequantizeLinear.eval(f8_values, scale_value, zero).flatten()
                q = onnx_ops.op_quantize_linear.QuantizeLinear_19.eval(values, scale_value, zero).flatten()
                qdq = onnx_ops.op_dequantize_linear.DequantizeLinear.eval(q, scale_value, zero).flatten()
                err_msg = (
                    f"{err_msg}\nvalues={values[:20]}\nqu={f8_values.flatten()[:20]}"
                    f"\n{q.flatten()[:20]}\ndq={dq[:20]}\nqdq={qdq[:20]}"
                )
            raise AssertionError(err_msg) from e


def check_model_correctness(
    testcase,
    model_path_origin,
    model_path_to_check,
    inputs,
    rtol=1e-2,
    atol=0.05,
    providers=None,
    dynamic=False,
    is_gemm=False,
    op_matmul=False,
):
    if providers is None:
        providers = ["CPUExecutionProvider"]
    sess_options = onnxruntime.SessionOptions()
    sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL
    sess_options.optimized_model_filepath = model_path_to_check + ".optimized.onnx"
    origin_sess = onnxruntime.InferenceSession(model_path_origin, sess_options=sess_options, providers=providers)
    origin_results = origin_sess.run(None, inputs)

    with open(model_path_origin, "rb") as f:
        model_onnx = onnx.load(f)
    ops_set = {node.op_type for node in model_onnx.graph.node}
    check_reference_evaluator = not (ops_set & {"EmbedLayerNormalization", "Conv", "Attention", "Transpose"})
    check_target_evaluator = False

    with open(model_path_to_check, "rb") as f:
        model_check = onnx.load(f)

    if check_reference_evaluator and onnx_recent_enough:
        ref = ReferenceEvaluator(model_path_origin)
        ref_origin_results = ref.run(None, inputs)
        for idx, ref_output in enumerate(origin_results):
            output = ref_origin_results[idx]
            np.testing.assert_allclose(
                ref_output,
                output,
                rtol=rtol,
                atol=atol,
                err_msg=f"Model {model_path_origin!r} failed for providers={providers!r}.",
            )

    # Verifies the shapes in the quantized model.
    if is_gemm:
        expected_shapes = {}
        with open(model_path_origin, "rb") as f:
            model = onnx.load(f)
            for init in model.graph.initializer:
                expected_shapes[init.name] = tuple(init.dims)
        checked = 0
        f8_quantization = False
        for init in model_check.graph.initializer:
            if init.name.endswith("_quantized"):
                name = init.name.replace("_quantized", "")
                expected = expected_shapes[name]
                shape = tuple(init.dims)
                if not dynamic and expected != shape:
                    raise AssertionError(
                        f"Shape mismatch for initializer {init.name!r} from {init.name!r}, "
                        f"shape={shape} != {expected} (expected)."
                    )
                else:
                    checked += 1
            if "zero_point" in init.name:
                dt = init.data_type
                f8_quantization = f8_quantization or dt in (
                    TensorProto.FLOAT8E4M3FN,
                    TensorProto.FLOAT8E4M3FNUZ,
                    TensorProto.FLOAT8E5M2,
                    TensorProto.FLOAT8E5M2FNUZ,
                )
        if checked == 0:
            raise AssertionError(
                f"Unable to check expected shape, expected_shapes={expected_shapes}, "
                f"names={[init.name for init in model_check.graph.initializer]}."
            )
        if f8_quantization:
            check_sign_f8_quantization(model_path_origin, model_path_to_check)

    # Verifies the expected outputs.
    if check_target_evaluator and onnx_recent_enough:
        if op_matmul:
            reference_new_ops = [QLinearMatMul]
        else:
            reference_new_ops = [QGemm]
        has_missing_reference_ops = any(
            node.domain not in ["", "ai.onnx"]
            and not any(
                node.domain == new_node.op_domain and node.op_type == new_node.__name__
                for new_node in reference_new_ops
            )
            for node in model_check.graph.node
        )
        if has_missing_reference_ops:
            # We need to skip the test if the model contains ops that are not supported.
            testcase.skipTest(
                f"Model {model_path_to_check!r} contains ops that are not supported by the reference evaluator."
            )
        # Needs pv.Version(onnx.__version__) >= pv.Version("1.16.0")
        ref = ReferenceEvaluator(model_check, new_ops=reference_new_ops)
        try:
            target_results = ref.run(None, inputs)
        except Exception as e:
            if "axis is out of boundary" not in str(e) and "list assignment index out of range" not in str(e):
                # Run through the same failure with more logs
                ref = ReferenceEvaluator(model_check, new_ops=reference_new_ops, verbose=10)
                target_results = ref.run(None, inputs)
            else:
                target_results = []
        if target_results:
            testcase.assertEqual(len(origin_results), len(target_results), "result count are different")
            for idx, ref_output in enumerate(origin_results):
                output = target_results[idx]
                np.testing.assert_allclose(
                    ref_output,
                    output,
                    rtol=rtol,
                    atol=atol,
                    err_msg=f"Model {model_path_to_check!r} failed for providers={providers!r}.",
                )

    # enable QDQ transformers
    # sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
    try:
        target_sess = onnxruntime.InferenceSession(
            model_path_to_check,
            sess_options=sess_options,
            providers=providers,
        )
    except (C.Fail, C.InvalidGraph) as e:
        # This should disabled when QDQ optimizers is implemented.
        se = str(e)
        if (
            "com.microsoft:QLinearMatMul(-1) is not a registered function/op" not in se
            and "Type 'tensor(float16)' of input parameter (input) of operator (QuantizeLinear)" not in se
            and "Type 'tensor(float16)' of input parameter (input) of operator (DynamicQuantizeLinear)" not in se
        ):
            # com.microsoft:QLinearMatMul is not yet implemented.
            # QuantizeLinear supports float16 in opset 19
            raise e
        return
    target_results = target_sess.run([], inputs)
    testcase.assertEqual(len(origin_results), len(target_results), "result count are different")
    for idx, ref_output in enumerate(origin_results):
        output = target_results[idx]
        np.testing.assert_allclose(
            ref_output,
            output,
            rtol=rtol,
            atol=atol,
            err_msg=f"Model {model_path_to_check!r} failed for providers={providers!r}.",
        )


def check_op_nodes(testcase, model_path, node_checker):
    model = onnx.load(Path(model_path))
    for node in model.graph.node:
        testcase.assertTrue(node_checker(node))


def check_qtype_by_node_type(testcase, model_to_check, check_list):
    if isinstance(model_to_check, str):
        model = onnx.load(model_to_check)
    elif isinstance(model_to_check, onnx.ModelProto):
        model = model_to_check
    # NOTE: ONNX shape inference does not work on MS domain nodes.
    # Therefore, this function cannot currently be used for graphs that contain ops such as
    # com.microsoft.QuantizeLinear, which support 16-bit quantization.
    model = onnx.shape_inference.infer_shapes(model)
    value_infos = {vi.name: vi for vi in model.graph.value_info}
    value_infos.update({ot.name: ot for ot in model.graph.output})
    value_infos.update({it.name: it for it in model.graph.input})
    initializers = {init.name: init for init in model.graph.initializer}

    for node in model.graph.node:
        if node.op_type in check_list:
            input_output_check_list = check_list[node.op_type]
            for check_item in input_output_check_list:
                tensor_name = node.input[check_item[1]] if check_item[0] == "i" else node.output[check_item[1]]
                if tensor_name not in value_infos and tensor_name not in initializers:
                    raise AssertionError(
                        f"Unable to find tensor_name={tensor_name!r} in {sorted(value_infos)}\n{model}"
                    )
                if tensor_name in value_infos:
                    vi = value_infos[tensor_name]
                    testcase.assertTrue(vi.type.HasField("tensor_type"))
                    testcase.assertTrue(vi.type.tensor_type.elem_type == check_item[2])
                else:  # if (tensor_name in initializers):
                    init = initializers[tensor_name]
                    testcase.assertEqual(init.data_type, check_item[2])


def create_clip_node(input_name, output_name, node_name, initializers, min_value=-1.0, max_value=1.0):
    clip_min_name = str(uuid.uuid4())
    clip_max_name = str(uuid.uuid4())
    clip_inputs = [input_name, clip_min_name, clip_max_name]
    clip_outputs = [output_name]
    clip_name = node_name
    initializers.append(onnx.numpy_helper.from_array(np.array(min_value, dtype=np.float32), name=clip_min_name))
    initializers.append(onnx.numpy_helper.from_array(np.array(max_value, dtype=np.float32), name=clip_max_name))
    return onnx.helper.make_node("Clip", clip_inputs, clip_outputs, name=clip_name)


def generate_random_initializer(initializer_name, tensor_shape, tensor_dtype, mean=0.0, dev=0.3):
    """
    Helper function to generate initializers for test inputs
    """
    tensor = np.random.normal(mean, dev, tensor_shape).astype(tensor_dtype)
    init = onnx.numpy_helper.from_array(tensor, initializer_name)
    return init


def get_tensor_consumers_and_producers(
    model: onnx.ModelProto,
) -> tuple[dict[str, list[onnx.NodeProto]], dict[str, onnx.NodeProto]]:
    """
    Returns a tuple containing the following python dictionaries:
      - consumers: maps a tensor name to the list of nodes that have that tensor as an input.
      - producers: maps a tensor name to the node that generates this tensor as an output.
    """
    consumers: dict[str, list[onnx.NodeProto]] = {}
    producers: dict[str, onnx.NodeProto] = {}
    for node in model.graph.node:
        # Iterate through node's inputs to build the consumers dictionary.
        for input_name in node.input:
            if input_name:
                if input_name not in consumers:
                    consumers[input_name] = []

                consumers[input_name].append(node)

        # Iterate through node's outputs to build the producers dictionary.
        for output_name in node.output:
            producers[output_name] = node

    return (consumers, producers)



import onnx
from onnx import TensorProto, helper


# This is to test the operators without "Qlinear" support but still support uint8 input
# These operators need to be internal to a graph/partition
# def GenerateModel(model_name):
def GenerateModel(model_name):  # noqa: N802
    nodes = [
        helper.make_node(
            "QuantizeLinear",
            ["X", "Scale", "Zero_point"],
            ["X_quantized"],
            "quantize_0",
        ),
        helper.make_node(
            "Concat",
            ["X_quantized", "X_quantized"],
            ["X_concat"],
            axis=-2,
            name="concat_0",
        ),
        helper.make_node(
            "MaxPool",
            ["X_concat"],
            ["X_maxpool"],
            kernel_shape=[2, 2],
            name="maxpool_0",
        ),
        helper.make_node(
            "Transpose",
            ["X_maxpool"],
            ["X_transposed"],
            perm=[0, 1, 3, 2],
            name="transpose_0",
        ),
        helper.make_node(
            "DequantizeLinear",
            ["X_transposed", "Scale", "Zero_point"],
            ["Y"],
            "dequantize_0",
        ),
    ]

    initializers = [
        helper.make_tensor("Scale", TensorProto.FLOAT, [1], [256.0]),
        helper.make_tensor("Zero_point", TensorProto.UINT8, [1], [0]),
    ]

    inputs = [
        helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 1, 1, 3]),
    ]

    graph = helper.make_graph(
        nodes,
        "NNAPI_Internal_uint8_Test",
        inputs,
        [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 1, 2, 1])],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


if __name__ == "__main__":
    GenerateModel("nnapi_internal_uint8_support.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import argparse
import logging
import os

import coloredlogs
from constants import (
    AttentionInputIDs,
    AttentionOutputIDs,
    MultiHeadAttentionInputIDs,
    MultiHeadAttentionOutputIDs,
    Operators,
)
from onnx import helper, load_model
from onnx_model import NodeProto, OnnxModel
from shape_infer_helper import SymbolicShapeInferenceHelper

logger = logging.getLogger(__name__)


class PackingAttentionBase:
    def __init__(self, model: OnnxModel, attention_op_type: str):
        self.model: OnnxModel = model
        self.nodes_to_remove: list = []
        self.nodes_to_add: list = []
        self.prune_graph: bool = False
        self.node_name_to_graph_name: dict = {}
        self.this_graph_name: str = self.model.model.graph.name
        self.attention_op_type = attention_op_type
        self.attention_nodes = self.model.get_nodes_by_op_type(attention_op_type)

    def _try_getting_attention_mask(self) -> str | None:
        mask_index = (
            AttentionInputIDs.MASK_INDEX
            if self.attention_op_type == Operators.ATTENTION
            else MultiHeadAttentionInputIDs.KEY_PADDING_MASK
        )
        first_attention_node = self._try_getting_first_attention()
        # check if attention has mask
        if not first_attention_node or len(first_attention_node.input) <= mask_index:
            return None

        attention_mask = first_attention_node.input[mask_index]

        # check if all attention nodes have same mask
        for node in self.attention_nodes:
            if len(node.input) <= mask_index or node.input[mask_index] != attention_mask:
                return None

        return attention_mask

    def _try_getting_first_attention(self) -> NodeProto | None:
        if len(self.attention_nodes) <= 0:
            return None

        return self.attention_nodes[0]

    def _try_getting_last_layernorm(self) -> NodeProto | None:
        last_layernorm_node = None
        for node in self.model.nodes():
            if node.op_type == Operators.LAYERNORM or node.op_type == Operators.SKIPLAYERNORM:
                last_layernorm_node = node
        return last_layernorm_node

    def _are_attentions_supported(self) -> bool:
        raise NotImplementedError()

    def _insert_removepadding_node(self, inputs: list[str], outputs: list[str]) -> None:
        new_node = helper.make_node(
            Operators.REMOVEPADDING,
            inputs=inputs,
            outputs=outputs,
            name=self.model.create_node_name(Operators.REMOVEPADDING),
        )

        new_node.domain = "com.microsoft"
        self.nodes_to_add.append(new_node)
        self.node_name_to_graph_name[new_node.name] = self.this_graph_name

    def _insert_restorepadding_node(self, inputs: list[str], outputs: list[str]) -> None:
        new_node = helper.make_node(
            Operators.RESTOREPADDING,
            inputs=inputs,
            outputs=outputs,
            name=self.model.create_node_name(Operators.RESTOREPADDING),
        )

        new_node.domain = "com.microsoft"
        self.nodes_to_add.append(new_node)
        self.node_name_to_graph_name[new_node.name] = self.this_graph_name

    def _replace_attention_with_packing_attention(self, token_offset: str, cumulative_sequence_length: str) -> None:
        raise NotImplementedError()

    def _get_input_to_remove_padding(self, first_attention_node) -> str | None:
        if self.attention_op_type == Operators.ATTENTION:
            return first_attention_node.input[AttentionInputIDs.INPUT]
        return None

    def convert(self, use_symbolic_shape_infer: bool = True) -> None:
        logger.debug("start converting to packing model...")

        if not self._are_attentions_supported():
            return

        attention_mask = self._try_getting_attention_mask()
        if not attention_mask:
            return

        first_attention_node = self._try_getting_first_attention()
        last_layernorm_node = self._try_getting_last_layernorm()
        if not last_layernorm_node:
            return

        # insert RemovePadding
        input_to_remove_padding = self._get_input_to_remove_padding(first_attention_node)
        if not input_to_remove_padding:
            return

        output_without_padding = input_to_remove_padding + "_no_padding"
        token_offset = input_to_remove_padding + "_token_offset"
        cumulated_seq_len = input_to_remove_padding + "_cumulated_seq_len"
        max_seq_len = input_to_remove_padding + "_max_seq_len"
        self._insert_removepadding_node(
            [input_to_remove_padding, attention_mask],
            [output_without_padding, token_offset, cumulated_seq_len, max_seq_len],
        )
        self.model.replace_input_of_all_nodes(input_to_remove_padding, output_without_padding)
        logger.debug("inserted RemovePadding before Attention")

        # insert RestorePadding
        restorepadding_input = last_layernorm_node.output[0] + "_restore_input"
        self._insert_restorepadding_node([restorepadding_input, token_offset], [last_layernorm_node.output[0]])
        self.model.replace_output_of_all_nodes(last_layernorm_node.output[0], restorepadding_input)
        logger.debug(f"inserted RestorePadding after last {last_layernorm_node.op_type} layer")

        # insert PackedAttention
        self._replace_attention_with_packing_attention(token_offset, cumulated_seq_len)
        logger.debug(f"replaced {self.attention_op_type} with Packed{self.attention_op_type}")

        self.model.remove_nodes(self.nodes_to_remove)
        self.model.add_nodes(self.nodes_to_add, self.node_name_to_graph_name)

        if self.prune_graph:
            self.model.prune_graph()
        elif self.nodes_to_remove or self.nodes_to_add:
            self.model.update_graph()
        self.model.clean_shape_infer()
        if use_symbolic_shape_infer:
            # Use symbolic shape inference since custom operators (like Gelu, SkipLayerNormalization etc)
            # are not recognized by onnx shape inference.
            shape_infer_helper = SymbolicShapeInferenceHelper(self.model.model, verbose=0)
            inferred_model = shape_infer_helper.infer_shapes(self.model.model, auto_merge=True, guess_output_rank=False)
            if inferred_model:
                self.model.model = inferred_model


class PackingAttention(PackingAttentionBase):
    def __init__(self, model: OnnxModel):
        super().__init__(model, Operators.ATTENTION)

    def _are_attentions_supported(self) -> bool:
        for node in self.attention_nodes:
            if OnnxModel.get_node_attribute(node, "past_present_share_buffer") is not None:
                return False
            if OnnxModel.get_node_attribute(node, "do_rotary") is not None:
                return False
            unidirection_attr = OnnxModel.get_node_attribute(node, "unidirectional")
            if unidirection_attr is not None and unidirection_attr != 0:
                return False
            if len(node.input) > AttentionInputIDs.PAST and not node.input[AttentionInputIDs.PAST]:
                return False
            if (
                len(node.input) > AttentionInputIDs.PAST_SEQUENCE_LENGTH
                and not node.input[AttentionInputIDs.PAST_SEQUENCE_LENGTH]
            ):
                return False
        return True

    def _replace_attention_with_packing_attention(self, token_offset: str, cumulative_sequence_length: str) -> None:
        for attention in self.attention_nodes:
            attention_bias = (
                attention.input[AttentionInputIDs.ATTENTION_BIAS]
                if len(attention.input) > AttentionInputIDs.ATTENTION_BIAS
                else ""
            )
            packed_attention = helper.make_node(
                Operators.PACKEDATTENTION,
                inputs=[
                    attention.input[AttentionInputIDs.INPUT],
                    attention.input[AttentionInputIDs.WEIGHTS],
                    attention.input[AttentionInputIDs.BIAS],
                    token_offset,
                    cumulative_sequence_length,
                    attention_bias,
                ],
                outputs=[attention.output[AttentionOutputIDs.OUTPUT]],
                name=self.model.create_node_name(Operators.PACKEDATTENTION),
            )

            attributes = []
            for attr in attention.attribute:
                if attr.name in ["num_heads", "qkv_hidden_sizes", "scale"]:
                    attributes.append(attr)

            packed_attention.attribute.extend(attributes)
            packed_attention.domain = "com.microsoft"
            self.nodes_to_add.append(packed_attention)
            self.nodes_to_remove.append(attention)
            self.node_name_to_graph_name[packed_attention.name] = self.this_graph_name

        logger.info("Converted %d Attention nodes to PackedAttention.", len(self.attention_nodes))


class PackingMultiHeadAttention(PackingAttentionBase):
    def __init__(self, model: OnnxModel):
        super().__init__(model, Operators.MULTI_HEAD_ATTENTION)

    def _check_empty_input(self, node, index: int, name: str):
        """Check a node does not have given input."""
        if len(node.input) > index:
            if len(node.input[index]) > 0:
                logger.error(f"node input {index} ({name}) is not supported in PackedMultiHeadAttention: {node}")
                return False
        return True

    def _check_empty_output(self, node, index: int, name: str):
        """Check a node does not have given input."""
        if len(node.output) > index:
            if len(node.output[index]) > 0:
                logger.error(f"node output {index} ({name}) is not supported in PackedMultiHeadAttention: {node}")
                return False
        return True

    def _are_attentions_supported(self) -> bool:
        for node in self.attention_nodes:
            for attr in node.attribute:
                if attr.name not in ["num_heads", "mask_filter_value", "scale"]:
                    logger.error(f"node attribute {attr.name} is not supported in PackedMultiHeadAttention: {node}")
                    return False

            if node.input[MultiHeadAttentionInputIDs.KEY] and not node.input[MultiHeadAttentionInputIDs.VALUE]:
                logger.error("packed kv format is not supported in PackedMultiHeadAttention")
                return False

            if not (
                self._check_empty_input(node, MultiHeadAttentionInputIDs.PAST_KEY, "past_key")
                and self._check_empty_input(node, MultiHeadAttentionInputIDs.PAST_VALUE, "past_key")
                and self._check_empty_output(node, MultiHeadAttentionOutputIDs.PRESENT_KEY, "present_key")
                and self._check_empty_output(node, MultiHeadAttentionOutputIDs.PRESENT_VALUE, "present_key")
            ):
                return False

        return True

    def _replace_attention_with_packing_attention(self, token_offset: str, cumulative_sequence_length: str) -> None:
        gated_relative_pos_bias_count = 0
        for mha in self.attention_nodes:
            attention_bias = (
                mha.input[MultiHeadAttentionInputIDs.ATTENTION_BIAS]
                if len(mha.input) > MultiHeadAttentionInputIDs.ATTENTION_BIAS
                else ""
            )
            packed_mha = helper.make_node(
                Operators.PACKED_MULTI_HEAD_ATTENTION,
                inputs=[
                    mha.input[MultiHeadAttentionInputIDs.QUERY],
                    mha.input[MultiHeadAttentionInputIDs.KEY],
                    mha.input[MultiHeadAttentionInputIDs.VALUE],
                    mha.input[MultiHeadAttentionInputIDs.BIAS],
                    token_offset,
                    cumulative_sequence_length,
                    attention_bias,
                ],
                outputs=[mha.output[MultiHeadAttentionOutputIDs.OUTPUT]],
                name=self.model.create_node_name(Operators.PACKED_MULTI_HEAD_ATTENTION),
            )

            attributes = []
            for attr in mha.attribute:
                if attr.name in ["num_heads", "mask_filter_value", "scale"]:
                    attributes.append(attr)

            packed_mha.attribute.extend(attributes)
            packed_mha.domain = "com.microsoft"
            self.nodes_to_add.append(packed_mha)
            self.nodes_to_remove.append(mha)
            self.node_name_to_graph_name[packed_mha.name] = self.this_graph_name

            # Append token_offset input to GatedRelativePositionBias
            if attention_bias:
                rel_pos_bias_node = self.model.get_parent(mha, MultiHeadAttentionInputIDs.ATTENTION_BIAS)
                if (
                    rel_pos_bias_node
                    and rel_pos_bias_node.op_type == "GatedRelativePositionBias"
                    and len(rel_pos_bias_node.input) == 6
                ):
                    rel_pos_bias_node.input.append(token_offset)
                    gated_relative_pos_bias_count += 1

        logger.info("Converted %d MultiHeadAttention nodes to PackedMultiHeadAttention.", len(self.attention_nodes))
        logger.info("Converted %d GatedRelativePositionBias nodes to packing mode.", gated_relative_pos_bias_count)

    def _get_input_to_remove_padding(self, first_attention_node) -> str | None:
        # When there are query, key and value inputs, we need to find the first input of the parent MatMul node.
        matmul = self.model.get_parent(first_attention_node, 0)
        if matmul and matmul.op_type == "MatMul":
            return matmul.input[0]
        return None


class PackingMode:
    def __init__(self, model: OnnxModel):
        self.model = model

    def convert(self, use_symbolic_shape_infer: bool = True) -> None:
        if self.model.get_nodes_by_op_type(Operators.ATTENTION):
            if self.model.get_nodes_by_op_type(Operators.MULTI_HEAD_ATTENTION):
                logger.error("Packing mode does not support both Attention and MultiHeadAttention in same graph.")
                return None
            packing = PackingAttention(self.model)
            return packing.convert(use_symbolic_shape_infer)
        elif self.model.get_nodes_by_op_type(Operators.MULTI_HEAD_ATTENTION):
            packing = PackingMultiHeadAttention(self.model)
            return packing.convert(use_symbolic_shape_infer)
        else:
            logger.error("Packing mode requires either Attention or MultiHeadAttention node in onnx graph.")
            return None


def _parse_arguments():
    parser = argparse.ArgumentParser(
        description="Convert to packing mode tool for ONNX Runtime. It converts BERT like model to use packing mode."
    )
    parser.add_argument("--input", required=True, type=str, help="input onnx model path")

    parser.add_argument("--output", required=True, type=str, help="optimized onnx model path")

    parser.add_argument("--verbose", required=False, action="store_true", help="show debug information.")
    parser.set_defaults(verbose=False)

    parser.add_argument(
        "--use_external_data_format",
        required=False,
        action="store_true",
        help="use external data format to store large model (>2GB)",
    )
    parser.set_defaults(use_external_data_format=False)

    args = parser.parse_args()

    return args


def _setup_logger(verbose):
    if verbose:
        coloredlogs.install(
            level="DEBUG",
            fmt="[%(filename)s:%(lineno)s - %(funcName)20s()] %(message)s",
        )
    else:
        coloredlogs.install(fmt="%(funcName)20s: %(message)s")


def main():
    args = _parse_arguments()

    _setup_logger(args.verbose)

    logger.debug(f"arguments:{args}")

    if os.path.realpath(args.input) == os.path.realpath(args.output):
        logger.warning("Specified the same input and output path. Note that this may overwrite the original model")

    model = load_model(args.input)
    packing_mode = PackingMode(OnnxModel(model))
    packing_mode.convert()
    packing_mode.model.save_model_to_file(args.output, use_external_data_format=args.use_external_data_format)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np
import onnx


def make_model(model_path: str):
    """
    Creates a QDQ model with a (DQ -> Transpose -> Q -> GRAPH OUTPUT) sequence. The Transpose is optimized out
    and the TransposeOptimizer should also remove the empty (DQ -> Q) sequence.
    """
    input0_shape = (1, 3, 4, 4)

    inputs = [onnx.helper.make_tensor_value_info("input0", onnx.TensorProto.FLOAT, input0_shape)]
    outputs = [onnx.helper.make_tensor_value_info("output0", onnx.TensorProto.UINT8, None)]

    mul_weight_scale_data = np.array(1.0, dtype=np.float32)
    mul_weight_zp_data = np.array(0, dtype=np.int8)

    initializers = [
        onnx.numpy_helper.from_array(np.array(1.0, dtype=np.float32), "scale_1"),
        onnx.numpy_helper.from_array(np.array(128, dtype=np.uint8), "zp_128"),
        onnx.numpy_helper.from_array(np.array(1.0 / 255.0, dtype=np.float32), "scale_inv_255"),
        onnx.numpy_helper.from_array(np.array(0, dtype=np.uint8), "zp_0"),
        onnx.numpy_helper.from_array(mul_weight_scale_data, "mul_weight_scale"),
        onnx.numpy_helper.from_array(mul_weight_zp_data, "mul_weight_zp"),
    ]
    nodes = []

    # Transpose to channel-last
    tp0_node = onnx.helper.make_node("Transpose", ["input0"], ["tp0_out"], name="tp0_node", perm=(0, 2, 3, 1))
    nodes.append(tp0_node)

    # Q_0
    q0_node = onnx.helper.make_node("QuantizeLinear", ["tp0_out", "scale_1", "zp_128"], ["q0_out"], name="q0_node")
    nodes.append(q0_node)

    # DQ_0
    dq0_node = onnx.helper.make_node("DequantizeLinear", ["q0_out", "scale_1", "zp_128"], ["dq0_out"], name="dq0_node")
    nodes.append(dq0_node)

    # Sigmoid
    sigmoid_node = onnx.helper.make_node("Sigmoid", ["dq0_out"], ["sigmoid_out"], name="sigmoid_node")
    nodes.append(sigmoid_node)

    # Q_1
    q1_node = onnx.helper.make_node(
        "QuantizeLinear", ["sigmoid_out", "scale_inv_255", "zp_0"], ["q1_out"], name="q1_node"
    )
    nodes.append(q1_node)

    # DQ_1
    dq1_node = onnx.helper.make_node(
        "DequantizeLinear", ["q1_out", "scale_inv_255", "zp_0"], ["dq1_out"], name="dq1_node"
    )
    nodes.append(dq1_node)

    # DQ for mul input[1]
    mul_weight_i8_data = np.array([1, 2, 3], dtype=np.int8)
    mul_weight = onnx.numpy_helper.from_array(mul_weight_i8_data, "mul_weight")
    initializers.append(mul_weight)

    nodes.append(
        onnx.helper.make_node(
            "DequantizeLinear",
            ["mul_weight", "mul_weight_scale", "mul_weight_zp"],
            ["mul_input_1"],
            name="dq_mul_input_1",
        )
    )

    # Mul
    mul_node = onnx.helper.make_node("Mul", ["dq1_out", "mul_input_1"], ["mul_out"], name="mul_node")
    nodes.append(mul_node)

    # Q_2
    q2_node = onnx.helper.make_node("QuantizeLinear", ["mul_out", "scale_inv_255", "zp_0"], ["q2_out"], name="q2_node")
    nodes.append(q2_node)

    # DQ_2
    dq2_node = onnx.helper.make_node(
        "DequantizeLinear", ["q2_out", "scale_inv_255", "zp_0"], ["dq2_out"], name="dq2_node"
    )
    nodes.append(dq2_node)

    # Transpose to channel-first
    tp1_node = onnx.helper.make_node("Transpose", ["dq2_out"], ["tp1_out"], name="tp1_node", perm=(0, 3, 1, 2))
    nodes.append(tp1_node)

    # Q_3 to graph output
    nodes.append(
        onnx.helper.make_node("QuantizeLinear", ["tp1_out", "scale_inv_255", "zp_0"], ["output0"], name="q3_node")
    )

    graph = onnx.helper.make_graph(
        nodes,
        "transpose_opt_empty_dqq_graph_output",
        inputs,
        outputs,
        initializer=initializers,
    )
    opset_imports = [
        onnx.helper.make_opsetid("", 19),
    ]
    qdq_model = onnx.helper.make_model(graph, opset_imports=opset_imports)

    print("[INFO]: Running onnx.checker on qdq model")
    qdq_model = onnx.shape_inference.infer_shapes(qdq_model)
    onnx.checker.check_model(qdq_model, True)

    print(f"[INFO]: Saving {model_path}")
    onnx.save_model(qdq_model, model_path)


if __name__ == "__main__":
    make_model("transpose_optimizer_empty_dq_q_at_graph_output.onnx")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# This script helps converting .npz files to .onnx_adapter files

import argparse
import os
import sys

import numpy as np

import onnxruntime as ort


def get_args() -> argparse:
    parser = argparse.ArgumentParser()
    parser.add_argument("--npz_file_path", type=str, required=True)
    parser.add_argument("--output_file_path", type=str, required=True)
    parser.add_argument("--adapter_version", type=int, required=True)
    parser.add_argument("--model_version", type=int, required=True)
    return parser.parse_args()


def export_lora_parameters(
    npz_file_path: os.PathLike, adapter_version: int, model_version: int, output_file_path: os.PathLike
):
    """The function converts lora parameters in npz to onnx_adapter format"""
    adapter_format = ort.AdapterFormat()
    adapter_format.set_adapter_version(adapter_version)
    adapter_format.set_model_version(model_version)
    name_to_ort_value = {}
    with np.load(npz_file_path) as data:
        for name, np_arr in data.items():
            ort_value = ort.OrtValue.ortvalue_from_numpy(np_arr)
            name_to_ort_value[name] = ort_value

    adapter_format.set_parameters(name_to_ort_value)
    adapter_format.export_adapter(output_file_path)


def main() -> int:
    args = get_args()
    export_lora_parameters(args.npz_file_path, args.adapter_version, args.model_version, args.output_file_path)
    return 0


if __name__ == "__main__":
    sys.exit(main())



import numpy as np
import onnx
from onnx import OperatorSetIdProto, TensorProto, helper, numpy_helper

batch = 3
hidden_size = 4
attention_head = 2
hidden_per_attention = 2

relative_attention_num_buckets = 32
input_len = 8
output_len = 8

X = helper.make_tensor_value_info("input", TensorProto.FLOAT, [batch, input_len, hidden_size])
Y = helper.make_tensor_value_info("output", TensorProto.FLOAT, [batch, output_len, hidden_size])

q_weight_np_vals = (0.01 * np.arange(hidden_size * hidden_size, dtype=np.float32)).reshape((hidden_size, hidden_size))
q_weight_initializer = numpy_helper.from_array(q_weight_np_vals, "encoder.layer.0.SelfAttention.q.weight")

k_weight_np_vals = (0.01 * np.arange(hidden_size * hidden_size, dtype=np.float32)).reshape((hidden_size, hidden_size))
k_weight_initializer = numpy_helper.from_array(k_weight_np_vals, "encoder.layer.0.SelfAttention.k.weight")

v_weight_np_vals = (0.01 * np.arange(hidden_size * hidden_size, dtype=np.float32)).reshape((hidden_size, hidden_size))
v_weight_initializer = numpy_helper.from_array(v_weight_np_vals, "encoder.layer.0.SelfAttention.v.weight")

q_bias_np_vals = 0.01 * np.arange(hidden_size, dtype=np.float32)
q_bias_initializer = numpy_helper.from_array(q_bias_np_vals, "encoder.layer.0.SelfAttention.q.bias")

k_bias_np_vals = 0.01 * np.arange(hidden_size, dtype=np.float32)
k_bias_initializer = numpy_helper.from_array(k_bias_np_vals, "encoder.layer.0.SelfAttention.k.bias")

v_bias_np_vals = 0.01 * np.arange(hidden_size, dtype=np.float32)
v_bias_initializer = numpy_helper.from_array(v_bias_np_vals, "encoder.layer.0.SelfAttention.v.bias")

q_starts_initializer = numpy_helper.from_array(np.asarray([0], dtype=np.int64), "q_starts")
k_starts_initializer = numpy_helper.from_array(np.asarray([hidden_size], dtype=np.int64), "k_starts")
v_starts_initializer = numpy_helper.from_array(np.asarray([2 * hidden_size], dtype=np.int64), "v_starts")

q_ends_initializer = numpy_helper.from_array(np.asarray([hidden_size], dtype=np.int64), "q_ends")
k_ends_initializer = numpy_helper.from_array(np.asarray([2 * hidden_size], dtype=np.int64), "k_ends")
v_ends_initializer = numpy_helper.from_array(np.asarray([9223372036854775807], dtype=np.int64), "v_ends")

slice_axes_initializer = numpy_helper.from_array(np.asarray([0], dtype=np.int64), "slice_axes")
slice_steps_initializer = numpy_helper.from_array(np.asarray([1], dtype=np.int64), "slice_steps")


transpose_q = helper.make_node(
    "Transpose",
    [q_weight_initializer.name],
    ["transpose_q"],
    name="transpose_q",
    perm=[1, 0],
)
transpose_k = helper.make_node(
    "Transpose",
    [k_weight_initializer.name],
    ["transpose_k"],
    name="transpose_k",
    perm=[1, 0],
)
transpose_v = helper.make_node(
    "Transpose",
    [v_weight_initializer.name],
    ["transpose_v"],
    name="transpose_v",
    perm=[1, 0],
)

matmul_q = helper.make_node("MatMul", ["input", "transpose_q"], ["matmul_q"], name="matmul_q")
matmul_k = helper.make_node("MatMul", ["input", "transpose_k"], ["matmul_k"], name="matmul_k")
matmul_v = helper.make_node("MatMul", ["input", "transpose_v"], ["matmul_v"], name="matmul_v")

concat_bias = helper.make_node(
    "Concat",
    [q_bias_initializer.name, k_bias_initializer.name, v_bias_initializer.name],
    ["concat_bias"],
    axis=0,
    name="concat_bias",
)

slice_q = helper.make_node(
    "Slice",
    [
        "concat_bias",
        q_starts_initializer.name,
        q_ends_initializer.name,
        slice_axes_initializer.name,
        slice_steps_initializer.name,
    ],
    ["slice_q"],
    name="slice_q",
)

slice_k = helper.make_node(
    "Slice",
    [
        "concat_bias",
        k_starts_initializer.name,
        k_ends_initializer.name,
        slice_axes_initializer.name,
        slice_steps_initializer.name,
    ],
    ["slice_k"],
    name="slice_k",
)

slice_v = helper.make_node(
    "Slice",
    [
        "concat_bias",
        v_starts_initializer.name,
        v_ends_initializer.name,
        slice_axes_initializer.name,
        slice_steps_initializer.name,
    ],
    ["slice_v"],
    name="slice_v",
)

add_q = helper.make_node("Add", ["matmul_q", "slice_q"], ["add_q"], name="add_q")
add_k = helper.make_node("Add", ["matmul_k", "slice_k"], ["add_k"], name="add_k")
add_v = helper.make_node("Add", ["matmul_v", "slice_v"], ["add_v"], name="add_v")

add_1 = helper.make_node("Add", ["add_q", "add_k"], ["add_1"], name="add_1")
add_2 = helper.make_node("Add", ["add_1", "add_v"], ["add_2"], name="add_2")
identity = helper.make_node("Identity", ["add_2"], ["output"], name="identity")

# Create the graph (GraphProto)
graph_def = helper.make_graph(
    [
        transpose_q,
        transpose_k,
        transpose_v,
        matmul_q,
        matmul_k,
        matmul_v,
        add_q,
        add_k,
        add_v,
        concat_bias,
        slice_q,
        slice_k,
        slice_v,
        add_1,
        add_2,
        identity,
    ],
    "concat-slice-test-model",
    [X],
    [Y],
    [
        q_weight_initializer,
        k_weight_initializer,
        v_weight_initializer,
        q_bias_initializer,
        k_bias_initializer,
        v_bias_initializer,
        q_starts_initializer,
        k_starts_initializer,
        v_starts_initializer,
        q_ends_initializer,
        k_ends_initializer,
        v_ends_initializer,
        slice_axes_initializer,
        slice_steps_initializer,
    ],
)

opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
onnxdomain.domain = ""  # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
opsets.append(onnxdomain)

msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"

opsets.append(msdomain)
kwargs = {}
kwargs["opset_imports"] = opsets

# Create the model (ModelProto)
model_def = helper.make_model(graph_def, producer_name="onnx-example", **kwargs)
onnx.save(model_def, "concat_slice_basic_test.onnx")



from enum import IntEnum, unique


class _OptimizerConfig:
    r"""Base class for optimizer configuration

    This class is not an optimizer, but a means to configure existing ones from ORT backend.
    Once configured, no user intervention is needed to update weights or zero gradients during training.
    The 'parameter group' concept described at :py:attr:`.params` is borrowed from
    `Pytorch <https://pytorch.org/docs/stable/optim.html#per-parameter-options>`_.

    Args:
        name (str): optimizer names.
            One of 'SGDOptimizer', 'AdamOptimizer' and 'LambOptimizer'
        defaults (dict): optimizer parameters applied to all model parameters.
                         Used when a parameter group doesn`t specify them.
                         NOTE: Every optimizer must have 'lr'.
        params (list of dict, default is []): list of parameter groups.
            Each dict must contain a 'params' key with a list of names of model's parameter that will
            be optimized with the group's custom hyper-parameters values.
            In other words, parameter groups override the default :py:attr:`.defaults`
            for specific model parameters.
            Empty list means all the parameters of the model will use :py:attr:`.defaults` during optimization.

    NOTE: To prevent model parameters to be trained, refer to :py:attr:`.ORTTrainerOptions.utils.frozen_weights`.

    Example:

    .. code-block:: python

        lamb_optim = _OptimizerConfig(name = 'LambOptimizer',
                                      params = [ { 'params' : ['model_param0', 'model_param1'],
                                                   'epsilon' : 0.03, 'beta' : 0.5},
                                                 { 'params' : ['model_param2'],
                                                   'alpha' : 0.04},
                                               ],
                                      defaults = { 'lr': 0.001, 'alpha' : 0.01, 'beta' : 0.9})
    """

    def __init__(self, name, params, defaults):
        assert isinstance(name, str), "'name' must be a string"
        assert name in [
            "AdamOptimizer",
            "LambOptimizer",
            "SGDOptimizer",
        ], "'name' must be one of 'AdamOptimizer', 'LambOptimizer' or 'SGDOptimizer'"
        assert isinstance(defaults, dict), "'defaults' must be a dict"
        assert "lr" in defaults, "'defaults' must contain a {'lr' : positive number} entry"
        assert (isinstance(defaults["lr"], float) or isinstance(defaults["lr"], int)) and defaults[  # noqa: SIM101
            "lr"
        ] >= 0, "lr must be a positive number"
        assert isinstance(params, list), "'params' must be a list"
        for group in params:
            assert isinstance(group, dict) and len(group) > 1 and "params" in group, (
                "Each dict inside 'params' must contain a {'params' : [model parameter names]} entry"
                " and additional entries for custom hyper parameter values"
            )
            for k in group:
                if k != "params":
                    assert k in defaults or k.replace("_coef", "") in defaults, (
                        f"'params' has {k} hyper parameter not present at 'defaults'"
                    )

        self.name = name
        self.lr = float(defaults["lr"])
        self.defaults = defaults
        self.params = []

        # TODO: monitor this for perf issues
        # Maybe we don't have to do this to populate TrainingParameters,
        # but it does make code easier to maintain
        for param_group in params:
            self._add_param_group(param_group)

    def _add_param_group(self, param_group):
        r"""Add a parameter group to the :py:class:`_OptimizerConfig` s `params`."""
        assert isinstance(param_group, dict), "param group must be a dict"

        # Each parameter group must have all hyper parameters set
        for name, value in self.defaults.items():
            if name not in param_group:
                param_group.setdefault(name, value)

        if "lambda_coef" in param_group:
            param_group["lambda"] = param_group.pop("lambda_coef")

        self.params.append(param_group)


class SGDConfig(_OptimizerConfig):
    r"""SGD optimizer configuration

    NOTE: Current implementation does not support :py:attr:`params`, and must be
    passed as an empty list.

    Args:
        params (list of dict, default is []): list of parameter groups.
            Each dict must contain a 'params' key with a list of names of model's parameter that will
            be optimized with the group's custom hyper-parameters values.
            In other words, parameter groups override the default :py:attr:`.defaults`
            for specific model parameters.
            Empty list means all the parameters of the model will use :py:attr:`.defaults` during optimization.
        lr (float, default is 0.001): Learning rate

    NOTE: To prevent model parameters to be trained, refer to :py:attr:`.ORTTrainerOptions.utils.frozen_weights`.

    Example:

    .. code-block:: python

        sgd_optim1 = SGDConfig(lr=0.001)
    """

    def __init__(self, params=[], lr=0.001):  # noqa: B006
        super().__init__(name="SGDOptimizer", params=params, defaults={"lr": lr})
        assert isinstance(params, list) and len(params) == 0, "'params' must be an empty list for SGD optimizer"


class AdamConfig(_OptimizerConfig):
    r"""Adam optimizer configuration

    Args:
        params (list of dict, default is []): list of parameter groups.
            Each dict must contain a 'params' key with a list of names of model's parameter that will
            be optimized with the group's custom hyper-parameters values.
            In other words, parameter groups override the default :py:attr:`.defaults`
            for specific model parameters.
            Empty list means all the parameters of the model will use :py:attr:`.defaults` during optimization.
        lr (float, default is 0.001): Learning rate
        alpha (float, default is 0.9): Coefficient of previous gradient in running average of row 1.
        beta (float, default is 0.999):  Coefficient of previous squared gradient in running average.
        lambda_coef (float, default is 0): Regularization coefficient.
        epsilon (float, default is 1e-8): Small scalar to avoid dividing by zero.
        do_bias_correction (bool, default is True): Compute unbiased 1st and 2nd momentums.
        weight_decay_mode (DecayMode, default is BEFORE_WEIGHT_UPDATE): Selects weight decay update strategy.

    NOTE: To prevent model parameters to be trained, refer to :py:attr:`.ORTTrainerOptions.utils.frozen_weights`.

    Example:

    .. code-block:: python

        # User-specified lr, alpha and weight_decay_mode for all model parameters
        adam_optim1 = AdamConfig(lr=0.01, alpha=0.85, weight_decay_mode=AdamConfig.DecayMode.AFTER_WEIGHT_UPDATE)

        # User-specified lr using parameters group
        adam_optim2 = AdamConfig([{'params':['fc1.weight','fc2.weight'], 'lr':0.005}], lr=0.01)
    """

    @unique
    class DecayMode(IntEnum):
        BEFORE_WEIGHT_UPDATE = (0,)
        AFTER_WEIGHT_UPDATE = 1

    def __init__(
        self,
        params=[],  # noqa: B006
        lr=0.001,
        alpha=0.9,
        beta=0.999,
        lambda_coef=0.0,
        epsilon=1e-8,
        max_norm_clip=1.0,
        do_bias_correction=True,
        weight_decay_mode=DecayMode.BEFORE_WEIGHT_UPDATE,
    ):
        assert lr >= 0, "'lr' must be a positive number"
        assert alpha >= 0, "'alpha' must be a positive number"
        assert beta >= 0, "'beta' must be a positive number"
        assert lambda_coef >= 0, "'lambda_coef' must be a positive number"
        assert epsilon >= 0, "'epsilon' must be a positive number"
        assert max_norm_clip != 0, "'max_norm_clip' must not be 0"
        assert isinstance(do_bias_correction, bool), "'do_bias_correction' must be a boolean"
        assert isinstance(weight_decay_mode, AdamConfig.DecayMode), "'weight_decay_mode' must be a AdamConfig.DecayMode"
        for param in params:
            assert "lr" not in param, "'lr' is not supported inside params"

        defaults = {
            "lr": lr,
            "alpha": alpha,
            "beta": beta,
            "lambda": lambda_coef,
            "epsilon": epsilon,
            "max_norm_clip": max_norm_clip,
            "do_bias_correction": do_bias_correction,
            "weight_decay_mode": weight_decay_mode,
        }
        super().__init__(name="AdamOptimizer", params=params, defaults=defaults)
        self.alpha = alpha
        self.beta = beta
        self.lambda_coef = lambda_coef
        self.epsilon = epsilon
        self.max_norm_clip = max_norm_clip
        self.do_bias_correction = do_bias_correction
        self.weight_decay_mode = weight_decay_mode


class LambConfig(_OptimizerConfig):
    r"""Lamb optimizer configuration

    Args:
        params (list of dict, default is []): list of parameter groups.
            Each dict must contain a 'params' key with a list of names of model's parameter that will
            be optimized with the group's custom hyper-parameters values.
            In other words, parameter groups override the default :py:attr:`.defaults`
            for specific model parameters.
            Empty list means all the parameters of the model will use :py:attr:`.defaults` during optimization.
        lr (float, default is 0.001): Learning rate
        alpha (float, default is 0.9): Coefficient of previous gradient in running average of row 1.
        beta (float, default is 0.999):  Coefficient of previous squared gradient in running average.
        lambda (float, default is 0): Regularization coefficient.
        ratio_min (float, default is -inf): Lower bound on confidence ratio.
        ratio_max (float, default is inf): Upper bound on confidence ratio.
        epsilon (float, default is 1e-6): Small scalar to avoid dividing by zero.
        do_bias_correction (bool, default is False): Compute unbiased 1st and 2nd momentums.

    NOTE: To prevent model parameters to be trained, refer to :py:attr:`.ORTTrainerOptions.utils.frozen_weights`.

    Example:

    .. code-block:: python

        # User-specified lr, alpha and weight_decay_mode for all model parameters
        lamb_optim1 = LambConfig(lr=0.01, alpha=0.85)

        # User-specified lr using parameters group
        lamb_optim2 = LambConfig([{'params':['fc1.weight','fc2.weight'], 'lr':0.005}], lr=0.01)
    """

    def __init__(
        self,
        params=[],  # noqa: B006
        lr=0.001,
        alpha=0.9,
        beta=0.999,
        lambda_coef=0.0,
        ratio_min=float("-inf"),
        ratio_max=float("inf"),
        epsilon=1e-6,
        max_norm_clip=1.0,
        do_bias_correction=False,
    ):
        assert lr >= 0, "'lr' must be a positive number"
        assert alpha >= 0, "'alpha' must be a positive number"
        assert beta >= 0, "'beta' must be a positive number"
        assert lambda_coef >= 0, "'lambda_coef' must be a positive number"
        assert isinstance(ratio_min, float), "'ratio_min' must be a valid float"
        assert isinstance(ratio_max, float), "'ratio_max' must be a valid float"
        assert epsilon >= 0, "'epsilon' must be a positive number"
        assert max_norm_clip != 0, "'max_norm_clip' must not be 0"
        assert isinstance(do_bias_correction, bool), "'do_bias_correction' must be a boolean"
        for param in params:
            assert "lr" not in param, "'lr' is not supported inside params"

        defaults = {
            "lr": lr,
            "alpha": alpha,
            "beta": beta,
            "lambda": lambda_coef,
            "ratio_min": ratio_min,
            "ratio_max": ratio_max,
            "epsilon": epsilon,
            "max_norm_clip": max_norm_clip,
            "do_bias_correction": do_bias_correction,
        }
        super().__init__(name="LambOptimizer", params=params, defaults=defaults)
        self.alpha = alpha
        self.beta = beta
        self.lambda_coef = lambda_coef
        self.ratio_min = ratio_min
        self.ratio_max = ratio_max
        self.epsilon = epsilon
        self.max_norm_clip = max_norm_clip
        self.do_bias_correction = do_bias_correction



#!/usr/bin/env python3

# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from __future__ import annotations

import argparse
import dataclasses
import json
import pathlib
import subprocess
import sys
import tempfile


def warn(message: str):
    print(f"WARNING: {message}", file=sys.stderr)


def parse_args():
    parser = argparse.ArgumentParser(
        description="Benchmark (https://github.com/google/benchmark) program runner. "
        "Runs a benchmark program until the benchmark measurements are within the desired coefficient of variation "
        "(CV) (stddev / mean) tolerance and outputs those measurements."
    )

    parser.add_argument(
        "--program",
        required=True,
        type=pathlib.Path,
        help="Path to the benchmark program to run.",
    )

    parser.add_argument(
        "--pattern",
        required=True,
        dest="patterns",
        action="extend",
        nargs="+",
        help="Benchmark test name pattern to specify which benchmark tests to run. "
        "Each pattern value will have its own invocation of the benchmark program (passed to the benchmark program "
        "with the --benchmark_filter option). "
        "To list the benchmark test names, run the benchmark program with the --benchmark_list_tests option.",
    )
    parser.add_argument(
        "--repetitions",
        type=int,
        default=10,
        help="Number of benchmark run repetitions (passed to the benchmark program with the "
        "--benchmark_repetitions option).",
    )

    parser.add_argument(
        "--max-cv",
        type=float,
        default=0.05,
        help="Maximum allowed CV (stddev / mean) value. "
        "The CV value is a number, not a percentage. E.g., a value of 0.05 corresponds to 5%%.",
    )
    parser.add_argument(
        "--max-attempts",
        type=int,
        default=3,
        help="Maximum number of times to attempt running the benchmark program.",
    )

    parser.add_argument(
        "--show-program-output",
        action="store_true",
        help="Display the output from the benchmark program.",
    )

    return parser.parse_args()


@dataclasses.dataclass
class BenchmarkResult:
    name: str
    median_real_time: float
    median_cpu_time: float
    time_unit: str


def run_benchmark(
    program: pathlib.Path,
    output_file: pathlib.Path,
    show_output: bool,
    pattern: str,
    repetitions: int,
    max_cv: float,
    max_attempts: int,
) -> list[BenchmarkResult]:
    benchmark_cmd = [
        f"{program}",
        f"--benchmark_filter={pattern}",
        f"--benchmark_repetitions={repetitions}",
        "--benchmark_report_aggregates_only",
        f"--benchmark_out={output_file}",
        "--benchmark_out_format=json",
    ]

    def check_cv(entries):
        valid = True

        for entry in entries:
            if entry.get("aggregate_name") != "cv":
                continue

            run_name = entry["run_name"]

            real_time_cv = float(entry["real_time"])
            if real_time_cv > max_cv:
                warn(f"real_time CV exceeds limit for run '{run_name}': {real_time_cv} > {max_cv}")
                valid = False

            cpu_time_cv = float(entry["cpu_time"])
            if cpu_time_cv > max_cv:
                warn(f"cpu_time CV exceeds limit for run '{run_name}': {cpu_time_cv} > {max_cv}")
                valid = False

        return valid

    def process_entries(entries) -> list[BenchmarkResult]:
        results = []

        for entry in entries:
            if entry.get("aggregate_name") != "median":
                continue

            result = BenchmarkResult(
                name=entry["run_name"],
                median_real_time=float(entry["real_time"]),
                median_cpu_time=float(entry["cpu_time"]),
                time_unit=entry["time_unit"],
            )

            results.append(result)

        return results

    attempts = 0
    while attempts < max_attempts:
        attempts += 1

        output_handle = None if show_output else subprocess.DEVNULL
        subprocess.run(
            benchmark_cmd,
            check=True,
            stdout=output_handle,
            stderr=output_handle,
            creationflags=subprocess.HIGH_PRIORITY_CLASS,
        )

        with open(output_file) as output:
            output_json = json.load(output)
            entries = output_json["benchmarks"]

        if not check_cv(entries):
            warn("Discarding benchmark run.")
            continue

        return process_entries(entries)

    raise RuntimeError("Failed to get measurements within the CV limit.")


def main():
    args = parse_args()

    program = args.program.resolve(strict=True)

    benchmark_results: list[BenchmarkResult] = []

    with tempfile.TemporaryDirectory() as temp_dir_name:
        temp_dir = pathlib.Path(temp_dir_name)
        output_file = temp_dir / "benchmark.out.json"

        for pattern in args.patterns:
            benchmark_results += run_benchmark(
                program=program,
                output_file=output_file,
                show_output=args.show_program_output,
                pattern=pattern,
                repetitions=args.repetitions,
                max_cv=args.max_cv,
                max_attempts=args.max_attempts,
            )

    print("name|median_real_time|median_cpu_time")
    print("-|-|-")
    for result in benchmark_results:
        print(
            f"{result.name}|"
            f"{round(result.median_real_time)} {result.time_unit}|"
            f"{round(result.median_cpu_time)} {result.time_unit}"
        )


if __name__ == "__main__":
    main()



from .config import AdamConfig, LambConfig, SGDConfig, _OptimizerConfig  # noqa: F401
from .fp16_optimizer import FP16_Optimizer  # noqa: F401
from .fused_adam import AdamWMode, FusedAdam  # noqa: F401
from .lr_scheduler import (
    ConstantWarmupLRScheduler,  # noqa: F401
    CosineWarmupLRScheduler,  # noqa: F401
    LinearWarmupLRScheduler,  # noqa: F401
    PolyWarmupLRScheduler,  # noqa: F401
    _LRScheduler,  # noqa: F401
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import os
import tempfile
from pathlib import Path

import numpy
import onnx
import torch
from io_binding_helper import TypeHelper
from models.t5.past_helper import PastKeyValuesHelper
from onnx_model import OnnxModel
from torch_onnx_export_helper import torch_onnx_export
from transformers import WhisperConfig, file_utils
from whisper_openai_helper import WhisperDecoderInitOpenai

from onnxruntime import InferenceSession

logger = logging.getLogger(__name__)


class WhisperDecoderInit(torch.nn.Module):
    """A Whisper decoder to create initial past key values.
    This model is only called once during starting decoding.
    """

    def __init__(
        self,
        decoder: torch.nn.Module,
        config: WhisperConfig,
        decoder_start_token_id: int | None = None,
    ):
        super().__init__()
        self.decoder = decoder
        self.config = config
        self.decoder_start_token_id = (
            decoder_start_token_id if decoder_start_token_id is not None else self.config.decoder_start_token_id
        )

    def forward(
        self,
        decoder_input_ids: torch.Tensor,
        encoder_hidden_states: torch.FloatTensor,
    ):
        encoder_outputs = file_utils.ModelOutput()
        encoder_outputs["last_hidden_state"] = encoder_hidden_states
        encoder_outputs["hidden_states"] = None
        encoder_outputs["attentions"] = None

        out = self.decoder.model(
            None,
            encoder_outputs=encoder_outputs,
            decoder_input_ids=decoder_input_ids,
            past_key_values=None,
            use_cache=True,
            return_dict=True,
        )
        logits = self.decoder.proj_out(out[0])
        return logits, out.past_key_values, out.encoder_last_hidden_state


class WhisperDecoder(torch.nn.Module):
    """A Whisper decoder with past key values"""

    def __init__(self, decoder, config, model_impl: str = "hf", model: torch.nn.Module = None):
        super().__init__()
        self.decoder = decoder
        self.config = config
        self.model_impl = model_impl
        if model is not None:
            self.whisper_decoder_openai_init = WhisperDecoderInitOpenai(model, decoder)

    def forward(self, decoder_input_ids, *past):
        encoder_outputs = file_utils.ModelOutput()
        dummy_encoder_hidden_states = torch.randn((decoder_input_ids.shape[0], 3000, int(self.config.d_model)))
        encoder_outputs["last_hidden_state"] = dummy_encoder_hidden_states
        encoder_outputs["hidden_states"] = dummy_encoder_hidden_states
        encoder_outputs["attentions"] = None

        if self.model_impl == "openai":
            dummy_encoder_hidden_states.unsqueeze(0)
            dec_out, present = self.whisper_decoder_openai_init(
                decoder_input_ids, dummy_encoder_hidden_states, past=past
            )
            return dec_out, present

        if len(past) == 0:
            past_key_values = None
        else:
            past_key_values = PastKeyValuesHelper.back_group_by_layer(past)

        decoder_out = self.decoder(
            None,
            encoder_outputs=encoder_outputs,
            decoder_input_ids=decoder_input_ids,
            past_key_values=past_key_values,
            use_cache=True,
            return_dict=True,
        )
        logits = decoder_out[0]
        present_self, _ = PastKeyValuesHelper.group_by_self_and_cross(decoder_out.past_key_values)
        return logits, present_self


class WhisperDecoderInputs:
    def __init__(
        self,
        decoder_input_ids,
        past_key_values=None,
    ):
        self.decoder_input_ids: torch.LongTensor = decoder_input_ids
        self.past_key_values: list[torch.FloatTensor] | list[torch.HalfTensor] | None = past_key_values

    @staticmethod
    def create_dummy(
        config: WhisperConfig,
        batch_size: int,
        encode_sequence_length: int,
        past_decode_sequence_length: int,
        device: torch.device,
        float16: bool = False,
        use_int32_inputs: bool = False,
        model_impl: str = "hf",
    ):  # -> WhisperDecoderInputs:
        """Create dummy inputs for WhisperDecoder.

        Args:
            decoder: decoder
            batch_size (int): batch size
            encode_sequence_length (int): sequence length of input_ids for encoder
            past_decode_sequence_length (int): past sequence length of input_ids for decoder
            device (torch.device): device of output tensors
            float16 (bool): whether the model uses float32 or float16 in input
            use_int32_inputs(bool): whether use int32 instead of int64 for some inputs

        Returns:
            WhisperDecoderInputs: dummy inputs for decoder
        """
        num_attention_heads: int = config.encoder_attention_heads
        num_layers: int = config.decoder_layers  # + config.encoder_layers
        vocab_size: int = config.vocab_size

        # Use head_size, use hidden_size / num_attention_heads here.
        # For example, whisper-large, d_model=1280 and num_heads=20
        head_size: int = config.d_model // config.encoder_attention_heads

        sequence_length: int = 1  # fixed for decoding
        decoder_input_ids = torch.randint(
            low=0,
            high=vocab_size - 1,
            size=(batch_size, sequence_length),
            dtype=(torch.int32 if use_int32_inputs else torch.int64),
            device=device,
        )

        float_type = torch.float16 if float16 else torch.float32

        if past_decode_sequence_length > 0:
            self_attention_past_shape = [
                batch_size,
                num_attention_heads,
                past_decode_sequence_length,
                head_size,
            ]
            cross_attention_past_shape = [
                batch_size,
                num_attention_heads,
                encode_sequence_length if model_impl == "hf" else past_decode_sequence_length,
                head_size,
            ]

            past = []
            for _ in range(2 * num_layers):
                past.append(torch.rand(self_attention_past_shape, dtype=float_type, device=device))

            for _ in range(2 * num_layers):
                past.append(torch.rand(cross_attention_past_shape, dtype=float_type, device=device))
        else:
            past = None

        return WhisperDecoderInputs(decoder_input_ids, past)

    def to_list(self) -> list:
        input_list = [self.decoder_input_ids]
        if self.past_key_values:
            input_list.extend(self.past_key_values)
        return input_list

    def to_fp32(self):
        past = [p.to(dtype=torch.float32) for p in self.past_key_values] if self.past_key_values else None
        return WhisperDecoderInputs(
            self.decoder_input_ids.clone(),
            past,
        )


class WhisperDecoderHelper:
    @staticmethod
    def export_onnx(
        decoder: WhisperDecoder,
        device: torch.device,
        onnx_model_path: str,
        verbose: bool = True,
        use_external_data_format: bool = False,
        use_int32_inputs: bool = False,
    ):
        """Export decoder to ONNX

        Args:
            decoder (Union[WhisperDecoder, WhisperDecoderNoPastState]): decoder object
            device (torch.device): device of decoder object
            onnx_model_path (str): onnx path
            verbose (bool, optional): print verbose information. Defaults to True.
            use_external_data_format (bool, optional): use external data format or not. Defaults to False.
            use_int32_inputs (bool, optional): use int32 inputs
        """
        assert isinstance(decoder, (WhisperDecoder, WhisperDecoderInit))

        inputs = WhisperDecoderInputs.create_dummy(
            decoder.config,
            batch_size=2,
            encode_sequence_length=3000,
            past_decode_sequence_length=6 if isinstance(decoder, WhisperDecoder) else 0,
            device=device,
            use_int32_inputs=use_int32_inputs,
            model_impl=decoder.model_impl,
        )
        input_list = inputs.to_list()

        # Fix past disappearing bug - duplicate first past entry
        # input_list.insert(2, input_list[2])

        past_names = PastKeyValuesHelper.get_past_names(decoder.config.decoder_layers, present=False)
        present_names = PastKeyValuesHelper.get_past_names(decoder.config.decoder_layers, present=True)
        present_self_names = present_names[: 2 * decoder.config.decoder_layers]

        input_past_names = past_names if isinstance(decoder, WhisperDecoder) else []
        output_present_names = present_self_names if isinstance(decoder, WhisperDecoder) else present_names
        output_names = ["logits", *output_present_names]

        # Shape of input tensors (sequence_length==1):
        #    input_ids: (batch_size, sequence_length)
        #    past_self_*: (batch_size, num_heads, past_decode_sequence_length, head_size)
        #    past_cross_*: (batch_size, num_heads, encode_sequence_length, head_size)

        # Shape of output tensors:
        #    logits: (batch_size, sequence_length, vocab_size)
        #    past_self_*: (batch_size, num_heads, past_decode_sequence_length + sequence_length, head_size)
        #    past_cross_*: (batch_size, num_heads, encode_sequence_length, head_size)

        input_names = ["input_ids"]
        input_names.extend(input_past_names)

        dynamic_axes = {
            "input_ids": {0: "batch_size"},
            "encoder_hidden_states": {0: "batch_size", 1: "encode_sequence_length / 2"},
            "logits": {0: "batch_size", 1: "sequence_length"},
        }

        for name in input_past_names:
            dynamic_axes[name] = {
                0: "batch_size",
                2: "past_decode_sequence_length" if "self" in name else "encode_sequence_length",
            }

        for name in output_present_names:
            if "cross" in name:
                dynamic_axes[name] = {0: "batch_size", 2: "encode_sequence_length"}
            else:  # self attention past state
                if isinstance(decoder, WhisperDecoder):
                    dynamic_axes[name] = {
                        0: "batch_size",
                        2: "past_decode_sequence_length + 1",
                    }
                else:
                    dynamic_axes[name] = {
                        0: "batch_size",
                        # 2: 'sequence_length'
                    }

        Path(onnx_model_path).parent.mkdir(parents=True, exist_ok=True)

        with tempfile.TemporaryDirectory() as tmp_dir_name:
            temp_onnx_model_path = os.path.join(tmp_dir_name, "decoder.onnx")
            Path(temp_onnx_model_path).parent.mkdir(parents=True, exist_ok=True)
            torch_onnx_export(
                decoder,
                args=tuple(input_list),
                f=temp_onnx_model_path if use_external_data_format else onnx_model_path,
                export_params=True,
                input_names=input_names,
                output_names=output_names,
                dynamic_axes=dynamic_axes,
                opset_version=17,
                do_constant_folding=True,
                use_external_data_format=use_external_data_format,
                verbose=verbose,
            )

            if use_external_data_format:
                model = onnx.load_model(temp_onnx_model_path, load_external_data=True)
                OnnxModel.save(
                    model,
                    onnx_model_path,
                    save_as_external_data=True,
                    all_tensors_to_one_file=True,
                )

    @staticmethod
    def onnxruntime_inference(ort_session, inputs: WhisperDecoderInputs):
        """Run inference of ONNX model."""
        logger.debug("start onnxruntime_inference")

        ort_inputs = {
            "input_ids": numpy.ascontiguousarray(inputs.decoder_input_ids.cpu().numpy()),
        }

        if inputs.past_key_values:
            assert len(inputs.past_key_values) % 4 == 0
            num_layers = int(len(inputs.past_key_values) / 4)
            past_names = PastKeyValuesHelper.get_past_names(num_layers)
            for i, past_tensor in enumerate(inputs.past_key_values):
                ort_inputs[past_names[i]] = numpy.ascontiguousarray(past_tensor.cpu().numpy())

        ort_outputs = ort_session.run(None, ort_inputs)
        return ort_outputs

    @staticmethod
    def verify_onnx(
        model: WhisperDecoder | WhisperDecoderInit,
        ort_session: InferenceSession,
        device: torch.device,
        use_int32_inputs: bool,
        max_cases: int = 4,
    ):
        """Compare the result from PyTorch and OnnxRuntime to verify the ONNX model is good."""
        float16: bool = TypeHelper.get_input_type(ort_session, "past_key_self_0") == "tensor(float16)"

        test_cases = [(4, 11, 3), (1, 2, 5), (3, 1, 1), (8, 5, 2)]
        test_cases_max_diff = []
        for (
            batch_size,
            encode_sequence_length,
            past_decode_sequence_length,
        ) in test_cases[:max_cases]:
            if isinstance(model, WhisperDecoderInit):
                dec_seq_len = 0
            else:
                dec_seq_len = past_decode_sequence_length

            inputs = WhisperDecoderInputs.create_dummy(
                model.config,
                batch_size,
                encode_sequence_length,
                dec_seq_len,
                device=device,
                float16=float16,
                use_int32_inputs=use_int32_inputs,
            )

            # We use fp32 PyTroch model as baseline even when ONNX model is fp16
            input_list = inputs.to_fp32().to_list()

            # Run inference of PyTorch model
            with torch.no_grad():
                torch_outputs = model(*input_list)

            ort_outputs = WhisperDecoderHelper.onnxruntime_inference(ort_session, inputs)

            max_diff = numpy.amax(numpy.abs(torch_outputs[0].cpu().numpy() - ort_outputs[0]))
            max_diff_all = max_diff
            logger.debug(f"logits max_diff={max_diff}")

            for i in range(2 * model.config.num_layers):
                max_diff = numpy.amax(numpy.abs(torch_outputs[1][i].cpu().numpy() - ort_outputs[1 + i]))
                logger.debug(f"self attention past state {i} max_diff={max_diff}")
                max_diff_all = max(max_diff_all, max_diff)

            if isinstance(model, WhisperDecoderInit):
                for i in range(2 * model.config.num_layers):
                    max_diff = numpy.amax(
                        numpy.abs(torch_outputs[2][i].cpu().numpy() - ort_outputs[1 + 2 * model.config.num_layers + i])
                    )
                    logger.debug(f"cross attention past state {i} max_diff={max_diff}")
                    max_diff_all = max(max_diff_all, max_diff)

            test_cases_max_diff.append(max_diff_all)
            logger.info(
                "batch_size=%s, encode_sequence_length=%s, past_decode_sequence_length=%s, max_diff=%s",
                batch_size,
                encode_sequence_length,
                past_decode_sequence_length,
                max_diff_all,
            )

        return max_diff_all



import os
import unittest

import onnxruntime_pybind11_state as C


class EPRegistrationTests(unittest.TestCase):
    def get_test_execution_provider_path(self):
        return os.path.join(".", "libtest_execution_provider.so")

    def test_register_custom_eps(self):
        C._register_provider_lib(
            "TestExecutionProvider", self.get_test_execution_provider_path(), {"some_config": "val"}
        )

        assert "TestExecutionProvider" in C.get_available_providers()

        this = os.path.dirname(__file__)
        custom_op_model = os.path.join(this, "testdata", "custom_execution_provider_library", "test_model.onnx")
        if not os.path.exists(custom_op_model):
            raise FileNotFoundError(f"Unable to find '{custom_op_model}'")

        session_options = C.get_default_session_options()
        sess = C.InferenceSession(session_options, custom_op_model, True, True)
        sess.initialize_session(["TestExecutionProvider"], [{"device_id": "0"}], set())
        print("Created session with customize execution provider successfully!")


if __name__ == "__main__":
    unittest.main()
